/*
Results: (epsilon = 81)
cond	--------- NEC-4D ------------- ------------- NEC-2D -----------
S/m	Rin 		Xin		Avg. Gain	Rin 		Xin 		Avg. Gain

0.001	7.69123E+01	1.26695E+02	0.160		7.90906E+01	1.29476E+02	0.155
0.003	7.21069E+01	1.25749E+02	0.167		7.38409E+01 	1.28334E+02 	0.162
0.01	6.06862E+01	1.19449E+02	0.218		6.02083E+01 	1.22146E+02	0.202
0.03	4.92272E+01	1.02255E+02	0.289		4.36457E+01	1.07793E+02 	0.323
0.1	4.10716E+01	7.86277E+01	0.409		2.84842E+01 	9.36058E+01 	0.583
0.3	2.78348E+01	6.67538E+01	0.664		1.96676E+01 	8.87374E+01 	0.929
1.0	-4.87995E+00	1.01535E+02	-4.060		1.59479E+01 	8.70547E+01 	1.218
5.0	-4.21415E+01	3.45929E+02	-0.492		.42931E+01 	8.59620E+01 	1.472
*/
void nec_context::rom2( nec_float a, nec_float b, complex_array& sum, nec_float dmin )
{
/*
      Z=A
      ZE=B
      S=B-A
      IF (S.GE.0.) GO TO 1
      WRITE(3,18)
      STOP
*/
	int nts = 4, nx = 1, n = 9;
	nec_float rx = 1e-4;
	nec_float dz = 0.0;
	nec_float dzot= dz*.5;

	nec_float _z = a;
	nec_float ze = b;
	nec_float _s = b - a;
	if ( _s < 0.0)
	{
		throw new nec_exception("ERROR - B LESS THAN A IN ROM2");
	}


/*
1     EP=S/(1.E4*NM)
      ZEND=ZE-EP
      DO 2 I=1,N
2     SUM(I)=(0.,0.)
      NS=NX
      NT=0
      CALL SFLDS (Z,G1)
*/
	complex_array g1(9), g2(9), g3(9), g4(9), g5(9);
	complex_array t01(9), t10(9), t20(9), t11(9);

	/*! tolerance for hitting upper limit */
	const nec_float ep = _s/(1.0e4 * m_geometry->n_plus_m);

	/*! upper limit */
	const nec_float zend = ze - ep;
	
	sum.fill(0,n,cplx_00());
	
	int ns = nx;
	/*! counter to control increasing subinterval size */
	int nt = 0;
	sflds( _z, g1);

/*
3     DZ=S/NS
      IF (Z+DZ.LE.ZE) GO TO 4
      DZ=ZE-Z
      IF (DZ.LE.EP) GO TO 17
4     DZOT=DZ*.5
      CALL SFLDS (Z+DZOT,G3)
      CALL SFLDS (Z+DZ,G5)
*/
loop3:	
	
	dz = _s / ns;
	if ( _z + dz >= ze)
	{
		dz = ze - _z;
		if ( dz <= ep)
			return;
	}

	dzot= dz*.5;
	sflds( _z + dzot, g3);
	sflds( _z + dz, g5);
	
	
/*
5     TMAG1=0.
      TMAG2=0.
C
C     EVALUATE 3 POINT ROMBERG RESULT AND TEST CONVERGENCE.
C
      DO 6 I=1,N
      T00=(G1(I)+G5(I))*DZOT
      T01(I)=(T00+DZ*G3(I))*.5
      T10(I)=(4.*T01(I)-T00)/3.
      IF (I.GT.3) GO TO 6
      TR=DREAL(T01(I))
      TI=DIMAG(T01(I))
      TMAG1=TMAG1+TR*TR+TI*TI
      TR=DREAL(T10(I))
      TI=DIMAG(T10(I))
      TMAG2=TMAG2+TR*TR+TI*TI
6     CONTINUE
*/
loop5:
	for (int i=0;i<n;i++)
	{
		nec_complex t00 = (g1[i]+ g5[i]) * dzot;
		t01[i] = (t00 + dz * g3[i]) * 0.5;
		t10[i] = (4.0 * t01[i] - t00)/3.;
	}

/*
      TMAG1=SQRT(TMAG1)
      TMAG2=SQRT(TMAG2)
      CALL TEST(TMAG1,TMAG2,TR,0.D0,0.D0,TI,DMIN)
*/
	nec_float tmag1 = sqrt( norm(t01[0]) + norm(t01[1]) + norm(t01[2]) );
	nec_float tmag2 = sqrt( norm(t10[0]) + norm(t10[1]) + norm(t10[2]) );

	nec_float tr = test_simple( tmag1, tmag2, dmin);

/*
      IF(TR.GT.RX)GO TO 8
      DO 7 I=1,N
7     SUM(I)=SUM(I)+T10(I)
      NT=NT+2
      GO TO 12
*/
	if (tr <= rx)
	{
		for (int i = 0; i < n; i++ )
			sum[i] += t10[i];
		nt += 2;
		goto loop12;
	}

/*
8     CALL SFLDS (Z+DZ*.25,G2)
      CALL SFLDS (Z+DZ*.75,G4)
      TMAG1=0.
      TMAG2=0.
C
C     EVALUATE 5 POINT ROMBERG RESULT AND TEST CONVERGENCE.
C
      DO 9 I=1,N
      T02=(T01(I)+DZOT*(G2(I)+G4(I)))*.5
      T11=(4.*T02-T01(I))/3.
      T20(I)=(16.*T11-T10(I))/15.
      IF (I.GT.3) GO TO 9
      TR=DREAL(T11)
      TI=DIMAG(T11)
      TMAG1=TMAG1+TR*TR+TI*TI
      TR=DREAL(T20(I))
      TI=DIMAG(T20(I))
      TMAG2=TMAG2+TR*TR+TI*TI
9     CONTINUE
      TMAG1=SQRT(TMAG1)
      TMAG2=SQRT(TMAG2)
      CALL TEST(TMAG1,TMAG2,TR,0.D0,0.D0,TI,DMIN)
*/
	sflds( _z+ dz*.25, g2);
	sflds( _z+ dz*.75, g4);
	
	/* Evaluate 5 point Romberg result and test convergence. */
	for (int i = 0; i < n; i++ )
	{
		nec_complex t02 = (t01[i]+ dzot*( g2[i]+ g4[i]))*0.5;
		t11[i] = (4.0 * t02- t01[i] )/3.0;
		t20[i] = (16.* t11[i] - t10[i])/15.0;
	}

	tmag1 = sqrt( norm(t11[0]) + norm(t11[1]) + norm(t11[2]) );
	tmag2 = sqrt( norm(t20[0]) + norm(t20[1]) + norm(t20[2]) );

	tr = test_simple( tmag1, tmag2, dmin);
/*
      IF(TR.GT.RX)GO TO 14
10    DO 11 I=1,N
11    SUM(I)=SUM(I)+T20(I)
      NT=NT+1
12    Z=Z+DZ
      IF (Z.GT.ZEND) GO TO 17
      DO 13 I=1,N
13    G1(I)=G5(I)
      IF (NT.LT.NTS.OR.NS.LE.NX) GO TO 3
      NS=NS/2
      NT=1
      GO TO 3
*/
	if (tr <= rx)
	{
loop10:
		for (int i = 0; i < n; i++ )
			sum[i] += t20[i];

		nt = nt+1;
	
loop12:
		_z = _z + dz;
		if ( _z > zend)
			return;
	
		for (int i = 0; i < n; i++ )
			g1[i] = g5[i];

		if ( (nt >= nts) && (ns > nx) )
		{
			/* Double step size */
			ns = ns/2;
			nt = 1;
		}
		goto loop3;
	}

/*
14    NT=0
      IF (NS.LT.NM) GO TO 15
      WRITE(3,19)  Z
      GO TO 10
15    NS=NS*2
      DZ=S/NS
      DZOT=DZ*.5
      DO 16 I=1,N
      G5(I)=G3(I)
16    G3(I)=G2(I)
      GO TO 5
*/
loop14:
	nt=0;
	if ( ns >= m_geometry->n_plus_m )
	{
		nec_error_mode em(m_output);
		m_output.endl();
		m_output.string("ROM2 -- STEP SIZE LIMITED AT Z = ");
		m_output.real_out(12,5,_z);
		goto loop10;
	}

	// halve step size 
	ns = ns*2;
	dz = _s/ ns;
	dzot = dz*.5;
	
	for (int i = 0; i < n; i++ )
	{
		g5[i] = g3[i];
		g3[i] = g2[i];
	}
	goto loop5;
}
/*
17    CONTINUE
      RETURN
C
18    FORMAT (30H ERROR - B LESS THAN A IN ROM2)
19    FORMAT (33H ROM2 -- STEP SIZE LIMITED AT Z =,1P,E12.5)
      END
*/